"""autocompletion command group"""

import click

from uncross.cli.uncross.autocompletion.bash import bash
from uncross.cli.uncross.autocompletion.fish import fish
from uncross.cli.uncross.autocompletion.zsh import zsh


@click.group("autocompletion")
def autocompletion() -> None:
    """Autocompletion commands."""


autocompletion.add_command(bash)
autocompletion.add_command(fish)
autocompletion.add_command(zsh)



"""project command group"""

import click

from uncross.cli.uncross.project.config.group import config


@click.group("project")
def project() -> None:
    """Project commands."""


project.add_command(config)



"""Run commands from cli perspective."""

from __future__ import annotations

import sys
from multiprocessing import Process

from uncross.cli import uncross
from uncross.logger import make_logger

LOGGER = make_logger(__name__)


def invoke_uncross(args: list[str]) -> None:
    """invoke uncross"""
    sys.argv = args
    uncross()  # pylint: disable=no-value-for-parameter


def run_uncross_process(args: list[str]) -> int:
    """run uncross test"""
    LOGGER.info("invoking: %s", args)
    process = Process(target=invoke_uncross, args=[args])
    process.start()
    process.join()
    return process.exitcode



"""CLI Entrypoint"""

import sys

from uncross.cli.uncross.group import uncross
from uncross.exceptions import FailedSubTaskError, ProgramMissingError, ToolchainMissingError
from uncross.logger import make_logger

LOGGER = make_logger(__name__)


def main() -> None:
    """Entrypoint for program"""
    try:
        uncross()  # pylint: disable=no-value-for-parameter
    except FailedSubTaskError:
        LOGGER.error("task failed, exiting ...")
        sys.exit(1)
    except (ProgramMissingError, ToolchainMissingError) as exc:
        LOGGER.error(str(exc))
        sys.exit(1)



"""Base pipeline class"""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from uncross.task.task import BuildTask


class BasePipeline:
    """Base pipeline class"""

    def __init__(self, name: str = "pipeline") -> None:
        self._name: str = name
        self.groups: dict[str, list[BuildTask]] = {}
        self.groups["base"] = []

    @property
    def name(self) -> str:
        """Get name"""
        return self._name

    def add_group(self, name: str, tasks: list[BuildTask]) -> None:
        """Register a build group"""
        if name in self.groups:
            raise RuntimeError
        self.groups[name] = tasks

    def add_task(self, task: BuildTask) -> None:
        """Add a base task"""
        self.groups["base"].append(task)

    def run(self) -> None:
        """Run the pipeline"""
        raise NotImplementedError



"""A build task to complete"""

from collections.abc import Callable


class BuildTask:
    """A build task"""

    def __init__(self, name: str, work: Callable) -> None:
        if not isinstance(name, str):
            raise TypeError
        if not isinstance(work, Callable):
            raise TypeError
        self._name = name
        self.work = work

    @property
    def name(self) -> str:
        """Get name"""
        return self._name

    def run(self) -> None:
        """Invoke work"""
        self.work()



"""Parse project config."""

from pathlib import Path

import toml

from uncross.git.repo import get_project_root
from uncross.logger import make_logger

LOGGER = make_logger(__name__)


def parse_project_config(search_path: str = ".") -> dict:
    """Parse project config"""
    config_path = Path(get_project_root(search_path=search_path)) / "uncross.toml"

    if not config_path.exists():
        return {}

    with config_path.open("r", encoding="utf-8") as config_file:
        config = toml.loads(config_file.read())
        LOGGER.debug("project config: %s", config)
        return config



"""Create a new project"""

from __future__ import annotations

import os
import pathlib

import rich.tree

from uncross.build_params import BuildParams
from uncross.cli.uncross.build import build_command
from uncross.git.repo import project_init_commit
from uncross.logger import make_logger
from uncross.project.artifacts.c_proj import create_c_project
from uncross.project.artifacts.clang_format import create_dot_clang_format
from uncross.project.artifacts.clang_tidy import create_dot_clang_tidy
from uncross.project.artifacts.cmake import create_cmakelists_txt
from uncross.project.artifacts.gitignore import create_dot_git_ignore
from uncross.project.artifacts.makefile import create_makefile
from uncross.project.artifacts.uncross import create_uncross_toml
from uncross.task.series_pipeline import SeriesPipeline

LOGGER = make_logger(__name__)


def display_project_overview_tree(source_dir: str) -> None:
    """Display project tree"""
    dir_style = "[magenta]"
    guide_style = "red"
    file_style = "[white]"

    def recurse_directory(directory: str, tree: rich.tree.Tree) -> None:
        """Recurse directory to fill project tree display"""

        paths: list[pathlib.Path] = sorted(
            pathlib.Path(directory).iterdir(), key=lambda p: (p.is_file(), p.name)
        )
        for path in paths:
            if path.is_dir():
                if path.name == "Debug":
                    debug_root = tree.add(f"{dir_style}{path.name}")
                    debug_root.add(f"{file_style}<debug build files>")
                elif path.name == "Release":
                    release_root = tree.add(f"{dir_style}{path.name}")
                    release_root.add(f"{file_style}<release build files>")
                elif path.name == ".git":
                    git_root = tree.add(f"{dir_style}{path.name}")
                    git_root.add(f"{file_style}<git files>")
                else:
                    recurse_directory(path, tree.add(f"{dir_style}{path.name}"))
            else:
                tree.add(f"{file_style}{path.name}")

    root = rich.tree.Tree(
        f"{dir_style}{source_dir}",
        guide_style=guide_style,
        hide_root=source_dir == pathlib.Path.cwd(),
    )
    recurse_directory(source_dir, root)
    rich.console.Console().print(root, markup=True)


def create_project(name: str, source_dir: str, build_project: bool, git: bool) -> None:
    """Create a builtool project tree"""

    if " " in name:
        msg = "Cannot have space in name"
        raise RuntimeError(msg)

    LOGGER.info("Creating project %s at %s...", name, source_dir)
    os.makedirs(source_dir, exist_ok=True)
    create_c_project(name, source_dir)
    create_dot_clang_format(name, source_dir)
    create_dot_clang_tidy(name, source_dir)
    create_cmakelists_txt(name, source_dir)
    create_dot_git_ignore(name, source_dir)
    create_makefile(name, source_dir)
    create_uncross_toml(name, source_dir)
    if build_project:
        params = BuildParams(
            build_dir=f"{source_dir}/build",
            source_dir=source_dir,
            build_debug=True,
            toolchains=["native"],
            presets=[],
            cmake_vars=[],
        )

        build_command(SeriesPipeline("build pipeline"), params)
        params.build_debug = False
        build_command(SeriesPipeline("build pipeline"), params)
    if git:
        project_init_commit(source_dir, name)
    display_project_overview_tree(source_dir)



"""download command"""

import tarfile
from pathlib import Path

import click

from uncross.logger import make_logger
from uncross.network.download import download_file

LOGGER = make_logger(__name__)


def download_bootlin_command(arch: str, libc: str, status: str, version: str) -> None:
    """download bootlin toolchain"""
    toolchain_base = f"{arch}--{libc}--{status}-{version}"
    tarball = f"{toolchain_base}.tar.bz2"
    disk_path = f"{Path.home()!s}/.uncross/cache/{tarball}"
    output_path = f"{Path.home()!s}/.uncross/toolchains/{toolchain_base}"

    if not Path(disk_path).exists():
        try:
            LOGGER.debug("Downloading %s to %s ...", tarball, disk_path)
            url = (
                "https://toolchains.bootlin.com/"
                "downloads/releases/toolchains/"
                f"{arch}/tarballs/{tarball}"
            )
            download_file(url, disk_path)
        except FileNotFoundError:
            LOGGER.error("toolchain %s not found.", tarball)
        except ConnectionError:
            LOGGER.error("Could not connect to toolchains.bootlin.com.")
    else:
        LOGGER.info("toolchain tarball found in cache")

    if not Path(output_path).exists():
        LOGGER.info("extracting ...")
        with tarfile.open(disk_path, mode="r:bz2") as tar:
            tar.extractall(path=output_path, filter=None)  # noqa: S202
    else:
        LOGGER.info("toolchain %s already extracted.", toolchain_base)


@click.command("bootlin")
@click.option("arch", "--arch", type=str, default="x86-64", help="Target architecture.")
@click.option("libc", "--libc", type=str, default="glibc", help="Target libc.")
@click.option(
    "status",
    "--status",
    type=str,
    default="stable",
    help="Target status ('stable', 'bleeding-edge').",
)
@click.option("version", "--version", type=str, default="2024.02-1", help="Target version.")
def bootlin(arch: str, libc: str, status: str, version: str):
    """Download from toolchains.bootlin.com."""
    download_bootlin_command(arch, libc, status, version)



"""toolchain register command group"""

import click

from uncross.cli.uncross.toolchain.register.bootlin import bootlin


@click.group("register")
def toolchain_register() -> None:
    """register toolchain with project"""


toolchain_register.add_command(bootlin)



"""Handle clang-format invocation"""

from __future__ import annotations

import functools

from uncross.invoke import (
    check_program,
    invoke_subprocess,
    perform_subtask,
    subtask_redirect_stdout,
)
from uncross.logger import make_logger

LOGGER = make_logger(__name__)


def invoke_clang_format(args: list[str], stdout_replace: str = "") -> int:
    """invoke clang format"""
    try:
        from clang_format import clang_format  # pylint: disable=import-outside-toplevel

        to_run = functools.partial(subtask_redirect_stdout, args, clang_format, stdout_replace)
    except ImportError:
        check_program("clang-format")
        task = functools.partial(invoke_subprocess, args)
        to_run = functools.partial(subtask_redirect_stdout, args, task, stdout_replace)

    return perform_subtask(to_run, [])



"""commit project config."""

from pathlib import Path

import toml

from uncross.git.repo import get_project_root
from uncross.logger import make_logger

LOGGER = make_logger(__name__)


def commit_project_config(document: dict, search_path: str = ".") -> None:
    """commit project config"""
    config_path = Path(get_project_root(search_path=search_path)) / "uncross.toml"

    with config_path.open("w", encoding="utf-8") as config_file:
        config: str = toml.dumps(document)
        LOGGER.debug("project config: %s", config)
        config_file.write(config)



"""Root cli group"""

import logging
import sys

import click

from uncross.__about__ import __version__
from uncross.cli.uncross.autocompletion.group import autocompletion
from uncross.cli.uncross.build import build
from uncross.cli.uncross.check import check
from uncross.cli.uncross.clean import clean
from uncross.cli.uncross.config.group import config
from uncross.cli.uncross.fmt import fmt
from uncross.cli.uncross.lint import lint
from uncross.cli.uncross.new import new
from uncross.cli.uncross.project.group import project
from uncross.cli.uncross.release import release
from uncross.cli.uncross.toolchain.group import toolchain
from uncross.logger import make_logger

LOGGER = make_logger(__name__)

ASCII_ART = (
    "__   _ _ __   ___ _ __ ___  ___ ___                                         \n"
    "| | | | '_ \\ / __| '__/ _ \\/ __/ __|                                        \n"
    "| |_| | | | | (__| | | (_) \\__ \\__ \\                                        \n"
    "\\___,_|_| |_|\\___|_|  \\___/|___/___/                                        \n"
)


@click.group(
    context_settings={"help_option_names": ["-h", "--help"]}, invoke_without_command=False
)
@click.version_option(version=__version__, prog_name="uncross")
@click.option("log_level", "--level", default="INFO", help="Logging level.")
def uncross(log_level: str):
    """Entrypoint."""

    try:
        log_format = "[%(levelname)s] %(message)s"
        if log_level == "DEBUG":
            log_format = "[%(levelname)s] %(filename)s.%(funcName)s:%(lineno)d %(message)s"
        logging.basicConfig(level=log_level.upper(), format=log_format)
    except ValueError:
        click.echo(f"ERROR: '{log_level}' is not a valid logging level.")
        sys.exit(1)


uncross.help = f"{ASCII_ART}\nAn opinionated meta build system for C cross-compilation."
uncross.add_command(autocompletion)
uncross.add_command(build)
uncross.add_command(check)
uncross.add_command(clean)
uncross.add_command(config)
uncross.add_command(fmt)
uncross.add_command(lint)
uncross.add_command(new)
uncross.add_command(project)
uncross.add_command(release)
uncross.add_command(toolchain)



"""autocompletion for fish"""

import click
from auto_click_auto import enable_click_shell_completion
from auto_click_auto.constants import ShellType


@click.command("fish")
def fish() -> None:
    """Add fish autocompletion."""
    enable_click_shell_completion(program_name="uncross", shells=[ShellType.FISH], verbose=True)



"""uncross config command group"""

import click

from uncross.cli.uncross.config.list import config_list


@click.group("config")
def config() -> None:
    """Configuration commands."""


config.add_command(config_list)



"""Build params."""

from __future__ import annotations

from dataclasses import dataclass


@dataclass
class BuildParams:
    """Parameter for building"""

    build_dir: str = "./build"
    source_dir: str = "."
    build_debug: bool = True
    cmake_vars: list[str] = ()
    toolchains: list[str] = ()
    presets: list[str] = ()



"""config list command"""

import click

from uncross.config.tool.parse import parse_tool_config
from uncross.logger import make_logger

LOGGER = make_logger(__name__)


@click.command("list")
def config_list() -> None:
    """New project."""
    config = parse_tool_config()

    LOGGER.info("config: %s", config)



"""Build command."""

from __future__ import annotations

import functools
import os
from multiprocessing import cpu_count
from typing import TYPE_CHECKING

import click

from uncross.build_params import BuildParams
from uncross.config.project.parse import parse_project_config
from uncross.exceptions import FailedSubTaskError, ToolchainMissingError
from uncross.git.repo import get_project_root
from uncross.logger import make_logger
from uncross.programs.cmake import invoke_cmake
from uncross.task.series_pipeline import SeriesPipeline
from uncross.task.task import BuildTask
from uncross.toolchains import get_toolchain_file_by_name

if TYPE_CHECKING:
    from pathlib import Path

    from uncross.task.base_pipeline import BasePipeline


LOGGER = make_logger(__name__)


def add_cmake_var_args(args: list, cmake_vars: list[str]) -> None:
    """Add cmake vars to arg list"""
    for var in cmake_vars:
        if "=" not in var:
            args.append(f"-D{var}=On")
        else:
            args.append(f"-D{var}")
    args.append("-DCMAKE_EXPORT_COMPILE_COMMANDS=On")


def task_configure_preset(name: str, preset: str, params: BuildParams) -> None:
    """Dummy task for testing"""
    LOGGER.debug("task: %s", name)

    build_mode = "Debug" if params.build_debug else "Release"
    build_dir = f"{params.build_dir}/{build_mode}/presets/{preset}"
    check_file = f"{build_dir}/CMakeCache.txt"
    if os.path.exists(check_file):
        LOGGER.debug("Found %s, skipping configure ...", check_file)
        return

    args = [
        "cmake",
        "--preset",
        preset,
        f"-B{params.build_dir}/{build_mode}/presets/{preset}",
        f"-S{params.source_dir}",
        f"-DCMAKE_BUILD_TYPE={build_mode}",
    ]

    add_cmake_var_args(args, params.cmake_vars)

    if invoke_cmake(args) != 0:
        raise FailedSubTaskError


def task_configure_toolchain(
    name: str, toolchain: str, params: BuildParams, toolchain_file: Path | None = None
) -> None:
    """Dummy task for testing"""
    LOGGER.debug("task: %s", name)

    build_mode = "Debug" if params.build_debug else "Release"
    build_dir = f"{params.build_dir}/{build_mode}/toolchains/{toolchain}"

    check_file = f"{build_dir}/CMakeCache.txt"
    if os.path.exists(check_file):
        LOGGER.debug("Found %s, skipping configure ...", check_file)
        return

    args = [
        "cmake",
        f"-B{build_dir}",
        f"-S{params.source_dir}",
        f"-DCMAKE_BUILD_TYPE={build_mode}",
    ]

    if toolchain_file is not None:
        args.append(f"-DCMAKE_TOOLCHAIN_FILE={toolchain_file!s}")

    add_cmake_var_args(args, params.cmake_vars)

    config = parse_project_config()

    if isinstance(
        config.get("uncross", {}).get("toolchain", {}).get(toolchain, {}).get("defines", None),
        dict,
    ):
        toolchain_vars = []
        for key, val in config["uncross"]["toolchain"][toolchain]["defines"].items():
            toolchain_vars.append(f"{key}={val}")
        add_cmake_var_args(args, toolchain_vars)

    if invoke_cmake(args) != 0:
        raise FailedSubTaskError


def task_build(name: str, build_subdir: str, build_name: str, params: BuildParams) -> None:
    """Build task"""
    LOGGER.debug("task: %s", name)

    build_mode = "Debug" if params.build_debug else "Release"
    args = [
        "cmake",
        "--build",
        f"{params.build_dir}/{build_mode}/{build_subdir}/{build_name}",
        "--parallel",
        f"{cpu_count() + 1}",
    ]

    if invoke_cmake(args) != 0:
        raise FailedSubTaskError


def build_preset(name: str, pipeline: BasePipeline, params: BuildParams) -> None:
    """Add build tasks for a preset"""
    configure_name = f"preset: {name} configure task"
    configure_work = functools.partial(task_configure_preset, configure_name, name, params)
    pipeline.add_task(BuildTask(configure_name, configure_work))
    build_name = f"preset: {name} build task"
    build_work = functools.partial(task_build, build_name, "presets", name, params)
    pipeline.add_task(BuildTask(build_name, build_work))


def build_toolchain(name: str, pipeline: BasePipeline, params: BuildParams) -> None:
    """Add build tasks for a toolchain"""
    configure_name = f"toolchain: {name} configure task"
    toolchain_file = None
    if name != "native":
        toolchain_file = get_toolchain_file_by_name(name)

        if toolchain_file is None:
            raise ToolchainMissingError(name)
    configure_work = functools.partial(
        task_configure_toolchain, configure_name, name, params, toolchain_file
    )
    pipeline.add_task(BuildTask(configure_name, configure_work))
    build_name = f"toolchain: {name} build task"
    build_work = functools.partial(task_build, build_name, "toolchains", name, params)
    pipeline.add_task(BuildTask(build_name, build_work))


def build_command(pipeline: BasePipeline, params: BuildParams) -> None:
    """Perform build."""
    LOGGER.info("building ...")

    for toolchain in params.toolchains:
        build_toolchain(toolchain, pipeline, params)

    for preset in params.presets:
        build_preset(preset, pipeline, params)

    pipeline.run()

    LOGGER.info("build complete")


@click.command("build")
@click.option("-S", "--source-dir", type=str, help="source directory")
@click.option("-B", "--build-dir", type=str, help="build directory")
@click.option("-D", "--define-cmake-var", type=str, multiple=True, help="define cmake variables")
@click.option("-t", "--toolchain", type=str, multiple=True, help="toolchain to invoke")
@click.option("-p", "--preset", type=str, multiple=True, help="cmake presets to build")
@click.option("--debug", is_flag=True, required=False, help="to build in debug")
@click.option("--release", is_flag=True, required=False, help="to build in release")
@click.option(
    "build_all",
    "--all",
    "-a",
    is_flag=True,
    required=False,
    help="to build both release and debug",
)
def build(
    source_dir: str,
    build_dir: str,
    define_cmake_var: list[str],
    toolchain: list[str],
    preset: list[str],
    debug: bool,
    release: bool,
    build_all: bool,
) -> None:
    """Build project."""
    LOGGER.debug("build command invoked with args:")

    if source_dir is None:
        source_dir = get_project_root()

    if build_dir is None:
        build_dir = f"{source_dir}/build"

    LOGGER.debug("source dir: %s", source_dir)
    LOGGER.debug("build dir: %s", build_dir)
    LOGGER.debug("define cmake variables: %s", define_cmake_var)
    LOGGER.debug("presets: %s", preset)

    config = parse_project_config(search_path=source_dir)

    if len(toolchain) == 0 and "uncross" in config and "toolchain" in config["uncross"]:
        toolchain = list(set(list(toolchain) + list(config["uncross"]["toolchain"].keys())))

    if len(toolchain) == 0 and len(preset) == 0:
        LOGGER.warning("no toolchains or presets provided, building native ...")
        toolchain = list(toolchain)
        toolchain.append("native")

    LOGGER.debug("toolchains: %s", toolchain)
    LOGGER.debug("debug: %s", debug)
    LOGGER.debug("release: %s", release)

    params = BuildParams(
        build_dir=build_dir,
        source_dir=source_dir,
        build_debug=False,
        toolchains=toolchain,
        presets=preset,
        cmake_vars=define_cmake_var,
    )

    if build_all or debug or (not debug and not release):
        LOGGER.debug("building Debug ...")
        params.build_debug = True
        build_command(SeriesPipeline("build pipeline"), params)

    if build_all or release:
        LOGGER.debug("building Release ...")
        params.build_debug = False
        build_command(SeriesPipeline("build pipeline"), params)



"""fmt command"""

from __future__ import annotations

import click

from uncross.cli.uncross.lint import run_over_c_h_files
from uncross.git.repo import get_project_root
from uncross.logger import make_logger

LOGGER = make_logger(__name__)


def fmt_command(source_dir: str | None) -> None:
    """fmt code"""
    LOGGER.info("formatting .c and .h files in %s", source_dir)
    args = ["clang-format", "-i"]
    if source_dir is None:
        args.append(f"--style=file:{get_project_root()}/.clang-format")
        format_failed = run_over_c_h_files(get_project_root(), args)
    else:
        args.append(f"--style=file:{source_dir}/.clang-format")
        format_failed = run_over_c_h_files(source_dir, args)

    if format_failed:
        raise RuntimeError


@click.command("fmt")
@click.option("-S", "--source-dir", type=str, help="source directory")
def fmt(source_dir: str) -> None:
    """Format project."""
    LOGGER.debug("fmt command invoked with args:")
    LOGGER.debug("source dir: %s", source_dir)
    fmt_command(source_dir)



"""Create cmake C Example Project Structure"""

import os

from uncross.logger import make_logger

LOGGER = make_logger(__name__)


def drop_root_cmakelists_txt(name: str, path: str) -> None:
    """Create C Source file"""

    root_cmakelists_txt_content = f"""
cmake_minimum_required(VERSION 3.13)

project({name} C)

add_subdirectory(src)

"""

    with open(path, "w", encoding="utf-8") as root_cmakelists_txt_file:
        root_cmakelists_txt_file.write(root_cmakelists_txt_content)


def drop_src_cmakelists_txt(name: str, path: str) -> None:
    """Create project header file"""

    src_cmakelists_txt_content = f"""
set(TARGET_NAME {name}_${{CMAKE_SYSTEM_NAME}}_${{CMAKE_SYSTEM_PROCESSOR}})

add_executable(${{TARGET_NAME}} main.c)
target_include_directories(${{TARGET_NAME}} PRIVATE ${{CMAKE_CURRENT_SOURCE_DIR}}/../include)

if (CMAKE_BUILD_TYPE STREQUAL "Debug")
  set_target_properties(
    ${{TARGET_NAME}}
    PROPERTIES
    RUNTIME_OUTPUT_DIRECTORY ${{CMAKE_SOURCE_DIR}}/debug/bin
  )
  target_compile_options(${{TARGET_NAME}} PRIVATE -O3 -g -Wall -Wextra -Wpedantic)
  if (${{CMAKE_SYSTEM_PROCESSOR}} STREQUAL "x86_64")
    target_compile_options(${{TARGET_NAME}} PRIVATE -fsanitize=address)
    target_link_options(${{TARGET_NAME}} PRIVATE -fsanitize=address)
  endif()
endif()

if (CMAKE_BUILD_TYPE STREQUAL "Release")
  set_target_properties(
    ${{TARGET_NAME}}
    PROPERTIES
    RUNTIME_OUTPUT_DIRECTORY ${{CMAKE_SOURCE_DIR}}/release/bin
  )
  target_compile_options(${{TARGET_NAME}} PRIVATE -O3 -Wall -Wextra -Wpedantic -s)
  target_link_options(${{TARGET_NAME}} PRIVATE -s)
endif()

"""

    with open(path, "w", encoding="utf-8") as src_cmakelists_txt_file:
        src_cmakelists_txt_file.write(src_cmakelists_txt_content)


def create_cmakelists_txt(name: str, root: str) -> None:
    """Create an example cmake C project dir structure"""
    LOGGER.debug("creating cmake C project structure at %s ...", root)
    os.makedirs(f"{root}/src", exist_ok=True)
    drop_root_cmakelists_txt(name, f"{root}/CMakeLists.txt")
    drop_src_cmakelists_txt(name, f"{root}/src/CMakeLists.txt")



"""autocompletion for bash"""

import click
from auto_click_auto import enable_click_shell_completion
from auto_click_auto.constants import ShellType


@click.command("bash")
def bash() -> None:
    """Add bash autocompletion."""
    enable_click_shell_completion(program_name="uncross", shells=[ShellType.BASH], verbose=True)



"""Check command."""

from __future__ import annotations

import functools
import webbrowser
from pathlib import Path
from typing import TYPE_CHECKING

import click

from uncross.build_params import BuildParams
from uncross.config.project.parse import parse_project_config
from uncross.exceptions import FailedSubTaskError
from uncross.git.repo import get_project_root
from uncross.logger import make_logger
from uncross.programs.code_checker import invoke_code_checker
from uncross.task.series_pipeline import SeriesPipeline
from uncross.task.task import BuildTask

if TYPE_CHECKING:
    from uncross.task.base_pipeline import BasePipeline

LOGGER = make_logger(__name__)


def task_check_toolchain(name: str, params: BuildParams) -> None:
    """Command:

    CodeChecker check
      -l <path to compile_commands.json>
      -o results
      --analyzer-config 'clang-tidy:take-config-from-directory=true'
    """
    build_mode = "Debug" if params.build_debug else "Release"
    analysis_path = f"{params.source_dir}/{build_mode.lower()}/analysis/code_checker/"
    build_dir = f"{params.build_dir}/{build_mode}/toolchains/{name}"

    compile_commands = f"{build_dir}/compile_commands.json"

    if not Path(compile_commands).exists():
        LOGGER.error("%s not found.", compile_commands)
        raise FailedSubTaskError

    LOGGER.info("running report on %s", compile_commands)
    args = [
        "CodeChecker",
        "check",
        "-l",
        compile_commands,
        "-o",
        f"{analysis_path}/results/{name}",
        "--analyzer-config",
        "'clang-tidy:take-config-from-directory=true'",
    ]

    result = invoke_code_checker(args)
    if result not in (0, 2):
        LOGGER.error("CodeChecker failed. (return code %s)", result)
        raise FailedSubTaskError


def task_report_toolchain(name: str, params: BuildParams, open_browser: bool) -> None:
    """Command:

    CodeChecker parse
      -e html
      ./results
      -o ./reports_html
    """
    build_mode = "Debug" if params.build_debug else "Release"
    analysis_path = f"{params.source_dir}/{build_mode.lower()}/analysis/code_checker/"

    results_dir = f"{analysis_path}/results/{name}"
    reports_dir = f"{analysis_path}/reports/{name}"

    args = [
        "CodeChecker",
        "parse",
        "-e",
        "html",
        results_dir,
        "-o",
        reports_dir,
    ]

    invoke_code_checker(args)

    if open_browser:
        report_path = f"{reports_dir}/index.html"
        LOGGER.info("opening %s in browser...", report_path)
        webbrowser.open_new_tab(report_path)


def check_toolchain(
    name: str, pipeline: BasePipeline, params: BuildParams, open_browser: bool
) -> None:
    """Add check tasks for a toolchain"""
    check_name = f"toolchain: {name} check task"
    check_work = functools.partial(task_check_toolchain, name, params)
    pipeline.add_task(BuildTask(check_name, check_work))
    report_name = f"toolchain: {name} reports task"
    report_work = functools.partial(task_report_toolchain, name, params, open_browser)
    pipeline.add_task(BuildTask(report_name, report_work))


def check_command(pipeline: BasePipeline, params: BuildParams, open_browser: bool = False) -> None:
    """Perform build."""
    LOGGER.info("checking ...")

    for toolchain in params.toolchains:
        check_toolchain(toolchain, pipeline, params, open_browser)

    for preset in params.presets:
        check_toolchain(preset, pipeline, params, open_browser)

    pipeline.run()

    LOGGER.info("check complete")


@click.command("check")
@click.option("-S", "--source-dir", type=str, help="source directory")
@click.option("-B", "--build-dir", type=str, help="build directory")
@click.option("-t", "--toolchain", type=str, multiple=True, help="toolchain to invoke")
@click.option("-p", "--preset", type=str, multiple=True, help="cmake presets to build")
@click.option("--debug", is_flag=True, required=False, help="to build in debug")
@click.option("--release", is_flag=True, required=False, help="to build in release")
@click.option(
    "check_all",
    "--all",
    "-a",
    is_flag=True,
    required=False,
    help="to check both release and debug",
)
@click.option(
    "open_browser", "--open", is_flag=True, required=False, help="open reports in browser"
)
def check(
    source_dir: str,
    build_dir: str,
    toolchain: list[str],
    preset: list[str],
    debug: bool,
    release: bool,
    check_all: bool,
    open_browser: bool,
) -> None:
    """Check project."""

    if source_dir is None:
        source_dir = get_project_root()

    if build_dir is None:
        build_dir = f"{source_dir}/build"

    LOGGER.debug("check command invoked with args:")
    LOGGER.debug("source dir: %s", source_dir)
    LOGGER.debug("build dir: %s", build_dir)
    LOGGER.debug("presets: %s", preset)

    config = parse_project_config(search_path=source_dir)

    if len(toolchain) == 0 and "uncross" in config and "toolchain" in config["uncross"]:
        toolchain = list(set(list(toolchain) + list(config["uncross"]["toolchain"].keys())))

    if len(toolchain) == 0 and len(preset) == 0:
        LOGGER.warning("no toolchains or presets provided, building native ...")
        toolchain = list(toolchain)
        toolchain.append("native")

    LOGGER.debug("toolchains: %s", toolchain)
    LOGGER.debug("debug: %s", debug)
    LOGGER.debug("release: %s", release)

    params = BuildParams(
        build_dir=build_dir,
        source_dir=source_dir,
        build_debug=False,
        toolchains=toolchain,
        presets=preset,
        cmake_vars=[],
    )

    if check_all or debug or (not debug and not release):
        LOGGER.debug("checking Debug ...")
        params.build_debug = True
        check_command(SeriesPipeline("build pipeline"), params, open_browser=open_browser)

    if check_all or release:
        LOGGER.debug("checking Release ...")
        params.build_debug = False
        check_command(SeriesPipeline("build pipeline"), params, open_browser=open_browser)



"""toolchain command group"""

import click

from uncross.cli.uncross.toolchain.download.group import toolchain_download
from uncross.cli.uncross.toolchain.list import toolchain_list
from uncross.cli.uncross.toolchain.register.group import toolchain_register


@click.group("toolchain")
def toolchain() -> None:
    """Toolchain management commands."""


toolchain.add_command(toolchain_download)
toolchain.add_command(toolchain_list)
toolchain.add_command(toolchain_register)



"""lint command"""

from __future__ import annotations

import os
import sys

import click
import git

from uncross.git.repo import get_project_root
from uncross.logger import make_logger
from uncross.programs.clang_format import invoke_clang_format

LOGGER = make_logger(__name__)


def run_over_c_h_files_find(source_dir: str | None, args: list[str]) -> bool:
    """Run over all files using a find-like behavior."""
    lint_found = False
    for dirpath, _, filenames in os.walk(source_dir):
        if "build" in dirpath or "__pycache__" in dirpath or "deps" in dirpath:
            continue
        for file in filenames:
            if file.endswith((".c", ".h")):
                LOGGER.debug("running clang-format on %s", file)

                file_args = [
                    os.path.join(dirpath, file),
                ]
                if invoke_clang_format(args + file_args) != 0:
                    lint_found = True

    return lint_found


def run_over_c_h_files_git(source_dir: str | None, args: list[str]) -> bool:
    """Run over all files by walking git tree."""
    lint_found = False

    repo = git.Repo(source_dir)

    staged_files = repo.index.diff("HEAD")
    staged_files = [item.a_path for item in staged_files if item.a_path.endswith((".c", ".h"))]

    committed_files = [
        item.abspath
        for item in repo.head.commit.tree.traverse()
        if item.abspath.endswith((".c", ".h"))
    ]

    for file in staged_files + committed_files:
        LOGGER.debug("running clang-format on %s", file)
        file_args = [file]
        if invoke_clang_format(args + file_args) != 0:
            lint_found = True

    return lint_found


def run_over_c_h_files(source_dir: str | None, args: list[str]) -> bool:
    """run for each c/h file"""
    try:
        return run_over_c_h_files_git(source_dir, args)
    except git.exc.InvalidGitRepositoryError:
        return run_over_c_h_files_find(".", args)


def lint_command(source_dir: str | None) -> None:
    """lint code"""
    LOGGER.info("linting .c and .h files ...")
    args = [
        "clang-format",
        "--dry-run",
        "-Werror",
    ]
    if source_dir is None:
        args.append(f"--style=file:{get_project_root()}/.clang-format")
        lint_found = run_over_c_h_files(get_project_root(), args)
    else:
        args.append(f"--style=file:{source_dir}/.clang-format")
        lint_found = run_over_c_h_files(source_dir, args)

    if lint_found:
        LOGGER.error("LINT FOUND")
        sys.exit(1)
    else:
        LOGGER.info("NO LINT FOUND")


@click.command("lint")
@click.option("-S", "--source-dir", type=str, help="source directory")
def lint(source_dir: str) -> None:
    """Lint code."""
    LOGGER.debug("lint command invoked with args:")
    LOGGER.debug("source dir: %s", source_dir)
    lint_command(source_dir)



"""Create .git_ignore"""

from uncross.logger import make_logger

LOGGER = make_logger(__name__)


def drop_git_ignore(path: str) -> None:
    """Drop git_ignore file"""

    git_ignore_contents = """
.vscode/
build/
release/
release.tar.gz
debug/
__pycache__/
dist/
*.o
*.a
*.so
"""
    with open(path, "w", encoding="utf-8") as git_ignore_file:
        git_ignore_file.write(git_ignore_contents)


def create_dot_git_ignore(_name: str, root: str) -> None:
    """Create .git_ignore"""
    LOGGER.debug("creating git_ignore file at %s ...", root)
    drop_git_ignore(f"{root}/.gitignore")



"""Toolchain management"""

from __future__ import annotations

import os
from pathlib import Path

from uncross.logger import make_logger

LOGGER = make_logger(__name__)


def enumerate_toolchains() -> dict[str, list]:
    """Get all available toolchains."""
    to_return: dict[str, list] = {}

    search_locations = (Path.home() / ".uncross" / "toolchains", Path("/opt/cross"))

    for location in search_locations:
        if not location.exists():
            continue

        path = Path(location)

        if not path.is_dir():
            continue

        to_return[str(location)] = []

        for toolchain in path.iterdir():
            if not toolchain.is_dir():
                continue
            to_return[str(location)].append(toolchain.name)

    return to_return


def find_toolchain_file(location: str, toolchain: str) -> Path | None:
    """find toolchain file given"""

    possibles = ("toolchainfile.cmake", "toolchain.cmake")

    path = Path(f"{location}/{toolchain}")
    LOGGER.debug("checking %s", path)

    if not path.exists():
        return None

    for loc, _, files in os.walk(str(path)):
        for item in files:
            if item in possibles:
                LOGGER.debug("found: %s", item)
                return path / loc / item

    return None


def get_toolchain_file_by_name(name: str) -> Path | None:
    """Get toolchain file by toolchain name"""
    toolchains = enumerate_toolchains()

    for location, possibles in toolchains.items():
        for toolchain in possibles:
            if toolchain == name:
                possible = find_toolchain_file(location, toolchain)
                if possible is None:
                    continue
                return possible

    return None



"""Testing project"""

from __future__ import annotations

import os
import secrets
import shutil
from pathlib import Path

from uncross.testing.run import run_uncross_process


class UncrossTestProject:
    """Project for testing."""

    def __init__(self, name: str, path: str) -> None:
        self.name = name
        self.path = Path(path)
        self._setup()

    def _setup(self) -> None:
        """Setup test project."""
        run_uncross_process(["uncross", "new", str(self.path), "--name", self.name])
        assert os.path.exists(str(self.path))
        assert os.path.exists(f"{self.path!s}/.clang-format")
        assert os.path.exists(f"{self.path!s}/.clang-tidy")
        assert os.path.exists(f"{self.path!s}/.gitignore")
        assert os.path.exists(f"{self.path!s}/CMakeLists.txt")
        assert os.path.exists(f"{self.path!s}/src/CMakeLists.txt")
        assert os.path.exists(f"{self.path!s}/src/main.c")
        assert os.path.exists(f"{self.path!s}/include/{self.name}.h")
        assert os.path.exists(f"{self.path!s}/uncross.toml")

    def build(self) -> None:
        """Build project"""
        run_uncross_process(
            ["uncross", "build", "-S", str(self.path), "-B", f"{self.path!s}/build", "--all"]
        )
        assert os.path.exists(f"{self.path}/build")
        assert os.path.exists(f"{self.path}/release/bin/{self.name}_Linux_x86_64")
        assert os.path.exists(f"{self.path}/debug/bin/{self.name}_Linux_x86_64")

    def teardown(self) -> None:
        """Teardown project."""
        shutil.rmtree(self.path, ignore_errors=True)


def project_vars() -> tuple[str, str, str]:
    """Vars to seed test project"""
    test_id = secrets.token_hex(12)
    project_name = "testproject"
    test_src_directory = f"/tmp/test_uncross/{test_id}"  # noqa: S108

    return test_id, project_name, test_src_directory



"""Pipeline that runs all tasks in series."""

from uncross.logger import make_logger
from uncross.task.base_pipeline import BasePipeline

LOGGER = make_logger(__name__)


class SeriesPipeline(BasePipeline):
    """Pipeline that runs all tasks in series."""

    def run(self) -> None:
        """do the thing."""

        if "base" not in self.groups:
            raise RuntimeError

        if len(self.groups["base"]) != 0:
            for task in self.groups["base"]:
                LOGGER.debug("running base task %s ...", task.name)
                task.run()

        for name, group in self.groups.items():
            if name == "base":
                continue
            for task in group:
                LOGGER.debug("running group %s task %s ...", name, task.name)
                task.run()

        LOGGER.debug("all tasks done")



"""invoke a subprocess"""

from __future__ import annotations

import multiprocessing
import os
import shutil
import subprocess
import sys
from typing import Callable

from uncross.exceptions import ProgramMissingError
from uncross.logger import make_logger

LOGGER = make_logger(__name__)


def check_program(program: str) -> None:
    """Check if a program exists"""
    if shutil.which(program) is None:
        raise ProgramMissingError(program)


def subtask_redirect_stdout(args: list[str], task: Callable, stdout_replace: str = "") -> None:
    """Run clang format"""
    sys.argv = args

    if stdout_replace != "":
        with open(stdout_replace, "w", encoding="utf-8") as output_path:
            os.dup2(output_path.fileno(), 1)
            return task()
    else:
        return task()


def invoke_subprocess(args: list[str]) -> int:
    """invoke a subprocess"""
    check_program(args[0])
    args[0] = shutil.which(args[0])
    LOGGER.debug("running command: %s", args)
    process = subprocess.run(args, shell=False, check=False)
    return process.returncode


def perform_subtask(task: Callable, args: list[str]) -> int:
    """Run a subtask"""
    process = multiprocessing.Process(target=task, args=args)
    process.start()
    process.join()
    return process.exitcode



"""autocompletion for zsh"""

import click
from auto_click_auto import enable_click_shell_completion
from auto_click_auto.constants import ShellType


@click.command("zsh")
def zsh() -> None:
    """Add zsh autocompletion."""
    enable_click_shell_completion(program_name="uncross", shells=[ShellType.ZSH], verbose=True)



"""toolchain download command group"""

import click

from uncross.cli.uncross.toolchain.download.bootlin import bootlin


@click.group("download")
def toolchain_download() -> None:
    """download toolchains"""


toolchain_download.add_command(bootlin)



"""Create .clang_format"""

from uncross.logger import make_logger
from uncross.programs.clang_format import invoke_clang_format

LOGGER = make_logger(__name__)


def drop_clang_format(path: str) -> None:
    """Drop clang format file"""
    args = [
        "clang-format",
        "--style",
        "{ BasedOnStyle: Microsoft, ColumnLimit: 80, SortIncludes: false }",
        "--dump-config",
    ]

    if invoke_clang_format(args, stdout_replace=path) != 0:
        raise RuntimeError


def create_dot_clang_format(_name: str, root: str) -> None:
    """Create .clang_format"""
    LOGGER.debug("creating clang format file at %s ...", root)
    drop_clang_format(f"{root}/.clang-format")



"""Create .clang_tidy"""

from uncross.logger import make_logger
from uncross.programs.clang_tidy import invoke_clang_tidy

LOGGER = make_logger(__name__)


def drop_clang_tidy(path: str) -> None:
    """Drop clang tidy file"""
    args = [
        "clang-tidy",
        '--checks="*,-altera*,-llvm-libc*,-android*"',
        "--dump-config",
    ]

    if invoke_clang_tidy(args, stdout_replace=path) != 0:
        raise RuntimeError


def create_dot_clang_tidy(_name: str, root: str) -> None:
    """Create .clang_tidy"""
    LOGGER.debug("creating clang tidy file at %s ...", root)
    drop_clang_tidy(f"{root}/.clang-tidy")



"""Create uncross.toml"""

import toml

from uncross.logger import make_logger

LOGGER = make_logger(__name__)


def drop_uncross_toml(name, path: str) -> None:
    """Drop uncross.toml file"""
    document = {
        "uncross": {
            "project": {"name": name, "author": "John Doe", "email": "john.doe@example.com"},
            "toolchain": {"native": {}},
        }
    }

    with open(path, "w", encoding="utf-8") as config_file:
        config_file.write(toml.dumps(document))


def create_uncross_toml(name: str, root: str) -> None:
    """Create uncross.toml"""
    LOGGER.debug("creating uncross.toml file at %s ...", root)
    drop_uncross_toml(name, f"{root}/uncross.toml")



"""test uncross basic project"""

from __future__ import annotations

import os

from uncross.testing.project import UncrossTestProject, project_vars
from uncross.testing.run import run_uncross_process


def test_repo_bootlin_toolchain():
    """Test creating a repo and building binaries"""
    _, project_name, test_src_directory = project_vars()
    test_project = UncrossTestProject(project_name, test_src_directory)
    return_code = run_uncross_process(
        [
            "uncross",
            "--level",
            "DEBUG",
            "toolchain",
            "register",
            "bootlin",
            "-S",
            test_src_directory,
            "--arch",
            "armv5-eabi",
        ]
    )
    assert return_code == 0
    test_project.build()
    assert os.path.exists(f"{test_src_directory}/release/bin/{project_name}_Linux_armv5l")
    assert os.path.exists(f"{test_src_directory}/debug/bin/{project_name}_Linux_armv5l")

    test_project.teardown()



"""Handle clang-format invocation"""

from __future__ import annotations

import functools
import sys

from codechecker_common.cli import main as code_checker

from uncross.invoke import perform_subtask
from uncross.logger import make_logger

LOGGER = make_logger(__name__)


def run_code_checker(args: list[str]) -> None:
    """Run code checker"""
    sys.argv = args
    code_checker()


def invoke_code_checker(args: list[str]) -> int:
    """invoke clang format"""
    to_run = functools.partial(run_code_checker, args)
    return perform_subtask(to_run, [])



"""Create C Example Project Structure"""

import os

from uncross.logger import make_logger

LOGGER = make_logger(__name__)


def drop_c_source_file(name: str, path: str) -> None:
    """Create C Source file"""

    source_content = f"""
#include <stdio.h>

#include "{name.lower()}.h"

int main(void)
{{
    DPRINTF("Project %s, debug mode\\n", "{name}");
    puts("Hello, {name}!");
    return 0;
}}

"""

    with open(path, "w", encoding="utf-8") as source_file:
        source_file.write(source_content)


def drop_c_header(name: str, path: str) -> None:
    """Create project header file"""

    header_guard = f"{name.upper()}_H_"
    header_content = f"""
#ifndef {header_guard}
#define {header_guard}

#include <stdio.h>

#ifndef NDEBUG
#define DPRINTF(fmt, ...) \
        do \
        {{ \
                printf("[DEBUG] "fmt, __VA_ARGS__); \
        }} while (0)
#define DEPUTS(str) DPRINTF("%s", str)
#else

#define DPRINTF(fmt, ...) do {{}} while(0)
#define DPUTS(str) do {{}} while(0)

#endif /* NDEBUG */

#endif  /* {header_guard} */

"""

    with open(path, "w", encoding="utf-8") as header_file:
        header_file.write(header_content)


def create_c_project(name: str, root: str) -> None:
    """Create an example C project dir structure"""
    LOGGER.debug("creating C project structure at %s ...", root)
    os.makedirs(f"{root}/src", exist_ok=True)
    os.makedirs(f"{root}/include", exist_ok=True)
    drop_c_header(name, f"{root}/include/{name.lower()}.h")
    drop_c_source_file(name, f"{root}/src/main.c")



"""new command"""

from __future__ import annotations

import sys

import click

from uncross.logger import make_logger
from uncross.project.create import create_project

LOGGER = make_logger(__name__)


def new_command(name: str | None, source_dir: str, build_project: bool, git: bool) -> None:
    """new project"""
    LOGGER.debug("Creating new project %s at %s ...", name, source_dir)
    if name is None:
        name = source_dir
    create_project(name, source_dir, build_project, git)


@click.command("new")
@click.argument("source_dir", type=str)
@click.option("-n", "--name", type=str, help="project name")
@click.option("--build", is_flag=True, required=False, help="build project on creation")
@click.option("--no-git", is_flag=True, required=False, help="initialize git repo")
def new(source_dir: str, name: str | None, build: bool, no_git: bool) -> None:
    """New project."""
    if name == "":
        name = source_dir
    LOGGER.debug("new command invoked with args:")
    LOGGER.debug("source dir: %s", source_dir)
    LOGGER.debug("name: %s", source_dir)
    if (" " in source_dir or "." in source_dir or "/" in source_dir) and name is None:
        LOGGER.error("'.', ' ', or '/' in path and name was not provided")
        sys.exit(1)

    new_command(name, source_dir, build, not no_git)



"""Git repo helper functions"""

import os

import git

from uncross.__about__ import __version__
from uncross.logger import make_logger

LOGGER = make_logger(__name__)


def get_project_root(search_path: str = ".") -> str:
    """Get the root path of the current project"""
    try:
        repo = git.Repo(search_path, search_parent_directories=True)
        return os.path.realpath(repo.working_dir)
    except git.exc.InvalidGitRepositoryError:
        LOGGER.warning(
            "current directory is not in a git repo,"
            "uncross works best when the project is "
            "committed at the root of a git repo."
        )
        return os.path.realpath(os.curdir)


def project_init_commit(base_path: str, project_name: str) -> None:
    """Initialize git repo and make first commit"""
    LOGGER.debug("creating bare repo at %s ...", base_path)
    repo = git.Repo.init(base_path)

    LOGGER.debug("adding files to repo ...")
    for file in repo.untracked_files:
        LOGGER.debug("adding %s to repo ...", file)
        repo.index.add(file)

    commit_message = f"Project {project_name} initialized with uncross version {__version__}."

    LOGGER.debug("committing ...")
    repo.index.commit(commit_message)



"""test uncross basic project"""

from __future__ import annotations

from uncross.testing.project import UncrossTestProject, project_vars
from uncross.testing.run import run_uncross_process


def test_repo_basic_create_lint_fmt():
    """Test creating a repo and linting/formatting"""

    _, project_name, test_src_directory = project_vars()

    test_project = UncrossTestProject(project_name, test_src_directory)

    # project starts with lint, this should fail
    return_code = run_uncross_process(
        ["uncross", "--level", "DEBUG", "lint", "-S", test_src_directory]
    )
    assert return_code != 0

    return_code = run_uncross_process(
        ["uncross", "--level", "DEBUG", "fmt", "-S", test_src_directory]
    )
    assert return_code == 0

    return_code = run_uncross_process(
        ["uncross", "--level", "DEBUG", "lint", "-S", test_src_directory]
    )
    assert return_code == 0

    test_project.teardown()



"""config list command"""

import click
import toml
from rich.console import Console
from rich.pretty import pprint
from rich.syntax import Syntax

from uncross.config.project.parse import parse_project_config


@click.command("list")
@click.option("as_json", "--json", is_flag=True, help="display as json")
def config_list(as_json: bool) -> None:
    """New project."""
    config = parse_project_config()

    if as_json is False:
        Console().print(Syntax(toml.dumps(config), "toml", line_numbers=True))
    else:
        pprint(config)



"""test uncross basic project"""

from __future__ import annotations

from uncross.testing.project import UncrossTestProject, project_vars


def test_repo_basic_create_build():
    """Test creating a repo and building binaries"""
    _, project_name, test_src_directory = project_vars()
    test_project = UncrossTestProject(project_name, test_src_directory)
    test_project.build()
    test_project.teardown()



"""Uncross exceptions"""


class FailedSubTaskError(Exception):
    """Raise if a subtask/subprocess fails"""


class ProgramMissingError(Exception):
    """Raise if a missing program"""

    def __init__(self, program: str) -> None:
        super().__init__(f"missing program: {program}")


class ToolchainMissingError(Exception):
    """Raise if a missing program"""

    def __init__(self, toolchain: str) -> None:
        super().__init__(f"missing toolchain: {toolchain}")



"""register command"""

import click

from uncross.cli.uncross.toolchain.download.bootlin import download_bootlin_command
from uncross.config.project.commit import commit_project_config
from uncross.config.project.parse import parse_project_config
from uncross.logger import make_logger

LOGGER = make_logger(__name__)


def register_bootlin_command(
    arch: str, libc: str, status: str, version: str, search_path: str = "."
) -> None:
    """register bootlin toolchain"""
    toolchain_base = f"{arch}--{libc}--{status}-{version}"
    config = parse_project_config(search_path=search_path)

    if "uncross" not in config:
        config["uncross"] = {}

    if "toolchain" not in config["uncross"]:
        config["uncross"]["toolchain"] = {}

    toolchain = {
        "source": "bootlin",
        "arch": arch,
        "libc": libc,
        "status": status,
        "version": version,
    }

    config["uncross"]["toolchain"][toolchain_base] = toolchain

    download_bootlin_command(arch, libc, status, version)
    commit_project_config(config, search_path=search_path)

    LOGGER.info("registered toolchain %s", toolchain_base)


@click.command("bootlin")
@click.option("-S", "--source-dir", type=str, help="source directory")
@click.option("arch", "--arch", type=str, default="x86-64", help="Target architecture.")
@click.option("libc", "--libc", type=str, default="glibc", help="Target libc.")
@click.option(
    "status",
    "--status",
    type=str,
    default="stable",
    help="Target status ('stable', 'bleeding-edge').",
)
@click.option("version", "--version", type=str, default="2024.02-1", help="Target version.")
def bootlin(source_dir: str, arch: str, libc: str, status: str, version: str):
    """register from toolchains.bootlin.com."""
    register_bootlin_command(arch, libc, status, version, source_dir if source_dir else ".")



"""clean command"""

import contextlib
import os
import shutil

import click

from uncross.git.repo import get_project_root
from uncross.logger import make_logger

LOGGER = make_logger(__name__)


def clean_command(project_root: str, build_dir: str) -> None:
    """Clean project"""
    debug_dir = f"{project_root}/debug"
    release_dir = f"{project_root}/release"
    LOGGER.info("Cleaning project ...")
    LOGGER.debug("removing build dir: %s ...", build_dir)
    shutil.rmtree(build_dir, ignore_errors=True)
    LOGGER.debug("removing debug dir: %s ...", debug_dir)
    shutil.rmtree(debug_dir, ignore_errors=True)
    LOGGER.debug("removing release dir: %s ...", release_dir)
    shutil.rmtree(release_dir, ignore_errors=True)
    LOGGER.debug("removing release tarball ...")
    with contextlib.suppress(FileNotFoundError):
        os.remove("release.tar.gz")


@click.command("clean")
@click.option("source_dir", "-S", "--source-dir", type=str, help="source directory")
@click.option("build_dir", "-B", "--build-dir", type=str, help="build directory")
def clean(source_dir: str, build_dir: str):
    """Clean build artifacts."""
    if source_dir is None:
        source_dir = get_project_root()

    if build_dir is None:
        build_dir = f"{source_dir}/build"

    LOGGER.debug("clean command invoked")
    clean_command(source_dir, build_dir)



"""Uncross logging."""

import logging


def make_logger(name: str) -> logging.Logger:
    """Get Logger and add handler and formatting.

    Args:
        name (str): Module name.

    Returns:
        logging.Logger: Logger object.
    """
    return logging.getLogger(name)



"""test uncross basic project"""

from __future__ import annotations

import os

from uncross.testing.project import UncrossTestProject, project_vars
from uncross.testing.run import run_uncross_process


def test_repo_basic_check():
    """Test creating a repo and building binaries"""

    _, project_name, test_src_directory = project_vars()
    test_build_directory = f"{test_src_directory}/build"

    test_project = UncrossTestProject(project_name, test_src_directory)

    test_project.build()

    run_uncross_process(
        ["uncross", "check", "-S", test_src_directory, "-B", test_build_directory, "--all"]
    )
    assert os.path.exists(
        f"{test_src_directory}/release/analysis/code_checker/reports/native/index.html"
    )
    assert os.path.exists(
        f"{test_src_directory}/debug/analysis/code_checker/reports/native/index.html"
    )

    test_project.teardown()



"""Download helpers."""

from http import HTTPStatus
from pathlib import Path

import requests
from rich.progress import BarColumn, DownloadColumn, Progress, TextColumn


def download_file(url, dest_path):
    """Download url file to path."""
    dest_path = Path(dest_path)
    dest_path.parent.mkdir(parents=True, exist_ok=True)

    try:
        with Progress(
            TextColumn("[bold blue]{task.fields[filename]}", justify="right"),
            BarColumn(),
            DownloadColumn(),
            TextColumn("[progress.percentage]{task.percentage:>3.1f}%"),
            "•",
            TextColumn("[progress.elapsed]{task.elapsed:>.2f} sec"),
        ) as progress:
            response = requests.get(url, stream=True, timeout=30)

            if response.status_code == HTTPStatus.NOT_FOUND:
                raise FileNotFoundError

            total = int(response.headers.get("content-length", 0))
            task = progress.add_task("Download", filename=dest_path.name, total=total)

            with open(dest_path, "wb") as file:
                for data in response.iter_content(chunk_size=1024):
                    file.write(data)
                    progress.update(task, advance=len(data))
    except requests.exceptions.ConnectionError as exc:
        raise ConnectionError from exc



"""List toolchains"""

from pathlib import Path

import click
import rich
import rich.style

from uncross.toolchains import enumerate_toolchains, find_toolchain_file


def toolchain_list_command(show_toolchain_file: bool = False) -> None:
    """List all toolchains"""
    console = rich.console.Console(highlighter=None)
    toolchains = enumerate_toolchains()

    for location in toolchains:
        console.print(f"{location}:", style="purple")

        for toolchain in Path(location).iterdir():
            if not toolchain.is_dir():
                continue
            console.print(f" - {toolchain.name}", markup=False, style="magenta")
            if show_toolchain_file:
                toolchain_file = find_toolchain_file(location, toolchain.name)
                path = str(toolchain_file).replace(
                    f"{location}/{toolchain.name}/{toolchain.name}", ""
                )
                console.print(f"   - {path}", markup=False, style="green")

        console.print("")


@click.command("list")
@click.option(
    "show_toolchain_file",
    "--show-toolchain-files",
    is_flag=True,
    required=False,
    help="initialize git repo",
)
def toolchain_list(show_toolchain_file: bool) -> None:
    """list toolchains"""
    toolchain_list_command(show_toolchain_file=show_toolchain_file)



"""release command"""

import os
import sys
import tarfile

import click

from uncross.git.repo import get_project_root
from uncross.logger import make_logger

LOGGER = make_logger(__name__)


def compress_directory_to_tar_gz(directory_path, output_file):
    """compress target directory"""
    LOGGER.info("tar %s to %s", directory_path, output_file)
    with tarfile.open(output_file, "w:gz") as tar:
        tar.add(directory_path, arcname=os.path.basename(directory_path))


def release_command(source_dir: str) -> None:
    """release code"""
    LOGGER.info("release project at %s", source_dir)
    if not os.path.exists(f"{source_dir}/release"):
        LOGGER.error("release dir does not exist, please run build --release")
        sys.exit(1)
    compress_directory_to_tar_gz(f"{source_dir}/release", f"{source_dir}/release.tar.gz")


@click.command("release")
@click.option("-S", "--source-dir", type=str, help="source directory")
def release(source_dir: str) -> None:
    """Release project."""
    if source_dir is None:
        source_dir = get_project_root()

    LOGGER.debug("release command invoked with args:")
    LOGGER.debug("source dir: %s", source_dir)
    release_command(source_dir)
